/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 1a
   In this version, there is no validation, and everything looks great!
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState  }        -- domain

one sig Lock {  var flagRaised: set Thread  }                     -- system

pred initialState {
   # Thread = 2                                                   -- domain
   Progs.progState = Thread -> Uninterested                       -- domain
   no Lock.flagRaised                     }                       -- system

/*-------------------------------------------------------------------------
ACTIONS
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {            -- action initiated by domain
   t.(Progs.progState) = Uninterested  }
pred callLock [t: Thread] { 
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting  -- change prog state
   Lock.flagRaised' = Lock.flagRaised + t                     -- raise flag
}

pred lockReturnEnabled [t: Thread] {          -- action initiated by system
   Lock.flagRaised = t             }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   flagRaised' = flagRaised                               -- no flag change
   Progs.progState' = Progs.progState ++ t -> InCritSec --change prog state
}

pred unlockEnabled [t: Thread] {              -- action initiated by domain
   t.(Progs.progState) = InCritSec  }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   Progs.progState' = Progs.progState ++ t -> Uninterested  -- change state
   Lock.flagRaised' = Lock.flagRaised - t                     -- lower flag
}

/*-------------------------------------------------------------------------
TRACES
   An instance of a model is a single trace.  The predicate anyTrace is
intended to describe everything that can happen.
-------------------------------------------------------------------------*/

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t]  )  }

pred anyTrace {  initialState && always delta }
run anyTrace expect 1 

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Requirements should describe the domain as we would like it to behave
when the system is implemented and installed.
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking {  some t: Thread | 
   (callLockEnabled [t] || lockReturnEnabled [t] || unlockEnabled [t])  }
assert noDeadlocks {  anyTrace => always nonDeadlocking  }
check noDeadlocks                                               -- VERIFIED

pred nonStarvation { all t: Thread | 
   t.(Progs.progState) = Waiting => 
      eventually t.(Progs.progState) = InCritSec  }
assert noStarvation {  anyTrace => always nonStarvation  }
check noStarvation                                              -- VERIFIED
